"""Serialization utilities.

Supported codecs
================

* **json**    - json with UTF-8 encoding.
* **pickle**  - pickle with base64 encoding (not urlsafe).
* **msgpack**  - https://msgpack.org/
* **protobuf**  - https://developers.google.com/protocol-buffers/

Serialization by name
=====================

The :func:`dumps` function takes a codec name and the object to encode,
then returns bytes:

.. sourcecode:: pycon

    >>> s = dumps(obj, 'json')

For the reverse direction, the :func:`loads` function takes a codec
name and bytes to decode:

.. sourcecode:: pycon

    >>> obj = loads(s, 'json')
"""

import json
import pickle
from typing import Any

default_codec = 'json'


def dumps(obj: Any, codec: str = None) -> bytes:
    """Encode object into bytes."""
    codec = codec or default_codec

    if codec == 'json':
        return json.dumps(obj).encode()
    elif codec == 'pickle':
        return pickle.dumps(obj)


def loads(data: bytes, codec: str = None) -> Any:
    """Decode object from bytes."""
    codec = codec or default_codec

    if codec == 'json':
        return json.loads(data.decode())
    elif codec == 'pickle':
        return pickle.loads(data)



from __future__ import annotations

from typing import Iterable

from mode import Service


class BaseService(Service):
    """
    Service life cycle.

    When starting (await Service.start())
    - The on_first_start callback is called.
    - Service logs: "[Service] Starting...".
    - on_start callback is called.
    - All @Service.task background tasks are started (in definition order).
    - All child services added by add_dependency(), or on_init_dependencies()) are started.
    - Service logs: "[Service] Started".
    - The on_started callback is called.

    When stopping (await Service.stop())
    - Service logs; "[Service] Stopping...".
    - The on_stop() callback is called.
    - All child services are stopped, in reverse order.
    - All asyncio futures added by add_future() are cancelled in reverse order.
    - Service logs: "[Service] Stopped".
    - If Service.wait_for_shutdown = True, it will wait for the Service.set_shutdown() signal to be called.
    - All futures started by add_future() will be gathered (awaited).
    - The on_shutdown() callback is called.
    - The service logs: "[Service] Shutdown complete!".

    When restarting (await Service.restart())
    - The service is stopped (await service.stop()).
    - The __post_init__() callback is called again.
    - The service is started (await service.start()).
    """

    def __post_init__(self) -> None:
        """Additional user initialization."""
        ...

    async def on_first_start(self) -> None:
        """Service started for the first time in this process."""
        self.log.info(self.label + ' first starting')

    async def on_start(self) -> None:
        """Service is starting."""
        self.log.info(self.label + ' starting')

    def on_init_dependencies(self) -> Iterable[BaseService]:
        """Return list of service dependencies for this service."""
        return []

    async def on_started(self) -> None:
        """Service has started."""
        self.log.info(self.label + ' started')

    async def on_stop(self) -> None:
        """Service is being stopped/restarted."""
        self.log.info(self.label + ' stopping')

    async def on_shutdown(self) -> None:
        """Service is stopped and shutdown."""
        self.log.info(self.label + ' shutdown')

    async def on_restart(self) -> None:
        """Service is being restarted."""
        self.log.info(self.label + ' restarting')



from contextlib import contextmanager
from typing import ContextManager

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, Session

from coresvc.base.base_model import BaseORM


class DatabaseSql:
    def __init__(self, uri):
        self.engine = create_engine(uri, pool_pre_ping=True)
        self.session_maker = sessionmaker(autocommit=False, autoflush=False, bind=self.engine)

    def create_tables(self):
        BaseORM.metadata.create_all(self.engine)

    @contextmanager
    def session_context(self) -> ContextManager[Session]:
        """Provide a transactional scope around a series of operations."""
        session: Session = self.session_maker()
        try:
            yield session
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()



from typing import Generic, TypeVar, Type, Callable, Optional, List

from sqlalchemy.orm import Session

from coresvc.base.base_model import BaseORM

ORMType = TypeVar("ORMType", bound=BaseORM)


class BaseCrudSql(Generic[ORMType]):
    def __init__(self, session_context: Callable[[], Session], orm_class: Type[ORMType]):
        """
        Base CRUD with db session to Create, Read, Update, Delete
        """
        self.session_context = session_context
        self.orm_class = orm_class

    def create(self, orm_obj: ORMType) -> ORMType:
        with self.session_context() as session:
            session.add(orm_obj)
            session.commit()
            session.refresh(orm_obj)
            return orm_obj

    def read(self, _id: str) -> Optional[ORMType]:
        with self.session_context() as session:
            return session.query(self.orm_class).filter(self.orm_class.id == _id).first()

    def read_multi(self, skip: int = 0, limit: int = 100) -> List[ORMType]:
        with self.session_context() as session:
            return session.query(self.orm_class).offset(skip).limit(limit).all()

    def update(self):
        raise NotImplementedError

    def remove(self):
        raise NotImplementedError



from typing import Optional

import pydantic
from sqlalchemy import Column, String
from sqlalchemy.ext.declarative import as_declarative


@as_declarative()
class BaseORM:
    """
    Base object relational mapper for associating user-defined Python classes with database tables
    """
    id = Column(String(36), primary_key=True)


class BaseModel(pydantic.BaseModel):
    """
    Base model for data container, validation and mapping between ORM
    """
    id: Optional[str] = None

    class Config:
        orm_mode = True



import random
import string
from uuid import uuid4


def gen_id() -> str:
    return str(uuid4())


def gen_string(length=32) -> str:
    return "".join(random.choices(string.ascii_letters, k=length))


def gen_float() -> float:
    pass


def gen_int() -> int:
    pass



from functools import cached_property
from typing import Callable, Dict, Any

import faust
from faust import TopicT


class KafkaClient(faust.App):
    def __init__(self, app_id, broker='kafka://', store='rocksdb://', topic_to_model: dict = None):
        self.topic_to_model = topic_to_model or {}
        super().__init__(app_id, broker=broker, store=store)

    async def on_first_start(self) -> None:
        self.log.info(f'registered_topics {self.registered_topics}')

    @cached_property
    def registered_topics(self) -> Dict[str, TopicT]:
        topics = {}
        for key, val in self.topic_to_model.items():
            topics[key] = self.topic(key, value_type=val)
        return topics

    async def publish(self, topic_name: str, value: Any):
        if topic := self.registered_topics.get(topic_name, None):
            return await topic.send(value=value)
        raise KeyError('topic not registered')

    def subscribe(self, topic_name: str, callback: Callable):
        async def process(stream: faust.Stream):
            async for value in stream:
                await callback(value)

        if topic := self.registered_topics.get(topic_name, None):
            return self.agent(topic)(process)
        raise KeyError('topic not registered')



from typing import Optional, Any

from aioredis import Redis

from coresvc.base.base_service import BaseService
import aioredis

from coresvc.serializer import serializer


class RedisStore(BaseService):
    def __init__(self, url, **kwargs):
        self.url = url
        self._redis: Optional[Redis] = None
        super().__init__(**kwargs)

    async def on_start(self) -> None:
        self._redis = await aioredis.create_redis_pool(self.url)

    async def on_started(self) -> None:
        self.log.info(self.label + ' started')

    async def on_stop(self) -> None:
        if self._redis:
            self._redis.close()
            await self._redis.wait_closed()

    async def set(self, key: str, data: Any) -> None:
        if not self.started:
            raise RuntimeError(self.label + ' not started')

        value = serializer.dumps(data)
        await self._redis.set(key, value)

    async def get(self, key: str) -> str:
        if not self.started:
            raise RuntimeError(self.label + ' not started')

        value = await self._redis.get(key)
        return serializer.loads(value)



from functools import cached_property
from typing import Optional, Callable, Any
from coresvc.serializer import serializer
from aio_pika import connect_robust, Connection, Channel, Message, Exchange, ExchangeType, IncomingMessage
from coresvc.base.base_service import BaseService


class AMQPClient(BaseService):
    def __init__(self, amqp_url, **kwargs):
        self.amqp_url = amqp_url
        super().__init__(**kwargs)

    def on_init_dependencies(self):
        return [self.producer, self.consumer]

    async def publish(self, topic: str, msg: dict):
        await self.producer.publish(topic, msg)

    async def subscribe(self, topic: str, handler: Callable):
        await self.consumer.subscribe(topic, handler)

    @cached_property
    def producer(self):
        return Producer(self.amqp_url, loop=self.loop, beacon=self.beacon)

    @cached_property
    def consumer(self):
        return Consumer(self.amqp_url, loop=self.loop, beacon=self.beacon)


class Producer(BaseService):
    def __init__(self, url, **kwargs):
        self.url = url
        self._connection: Optional[Connection] = None
        self._channel: Optional[Channel] = None
        self._exchange: Optional[Exchange] = None
        super().__init__(**kwargs)

    async def on_start(self) -> None:
        await self._wait_for_connection()
        self._channel = await self._connection.channel()
        self._exchange = await self._channel.declare_exchange('pubsub', type=ExchangeType.TOPIC)
        # TODO: Consider if exchange name should be
        #  - a single name for all or
        #  - multiple ones consistent with service namespace
        #  - this also depends on if we need more than one type of exchanges: fanout, direct, topic, etc

    async def on_started(self) -> None:
        self.log.info(self.label + 'started')

    async def on_stop(self) -> None:
        if self._connection:
            await self._connection.close()

    async def publish(self, topic: str, msg: dict):
        if not self.started:
            raise RuntimeError('service not started')
        message = Message(body=serializer.dumps(msg))
        await self._exchange.publish(message, routing_key=topic)

    async def _wait_for_connection(self, retry=0):
        try:
            self._connection = await connect_robust(self.url, loop=self.loop)
        except ConnectionError:
            if retry < 5:
                await self.sleep(3)
                return await self._wait_for_connection(retry+1)
            raise


class Consumer(BaseService):
    def __init__(self, url, **kwargs):
        self.url = url
        self._connection: Optional[Connection] = None
        self._channel: Optional[Channel] = None
        self._exchange: Optional[Exchange] = None
        self._msg_handlers = {}
        super().__init__(**kwargs)

    async def on_start(self) -> None:
        await self._wait_for_connection()
        self._channel = await self._connection.channel()
        self._exchange = await self._channel.declare_exchange('pubsub', type=ExchangeType.TOPIC)

        # Maximum message count which will be
        # processing at the same time.
        # await self._channel.set_qos(prefetch_count=100)

    async def on_started(self) -> None:
        self.log.info(self.label + 'started')

    async def on_stop(self) -> None:
        if self._connection:
            await self._connection.close()

    async def subscribe(self, topic: str, handler: Callable[[dict], Any]):
        if not self.started:
            raise RuntimeError('service not started')

        async def handler_wrapper(message: IncomingMessage):
            obj = serializer.loads(message.body)
            try:
                await handler(obj)
                message.ack()  # msg handled properly and positive acknowledge
            except Exception as e:
                self.log.exception(str(e))
                message.nack()  # msg wasn't handled properly, negative acknowledge and requeue.

        queue = await self._channel.declare_queue()
        await queue.bind(self._exchange, routing_key=topic)
        await queue.consume(handler_wrapper)

    async def _wait_for_connection(self, retry=0):
        try:
            self._connection = await connect_robust(self.url, loop=self.loop)
        except ConnectionError:
            if retry < 5:
                await self.sleep(3)
                return await self._wait_for_connection(retry+1)
            raise



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['coresvc',
 'coresvc.base',
 'coresvc.comm',
 'coresvc.crud',
 'coresvc.datastore',
 'coresvc.serializer',
 'coresvc.util']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.3.22,<2.0.0',
 'aio-pika>=6.7.1,<7.0.0',
 'aioredis>=1.3.1,<2.0.0',
 'fastapi>=0.63.0,<0.64.0',
 'faust>=1.10.4,<2.0.0',
 'psycopg2>=2.8.6,<3.0.0',
 'pydantic>=1.7.3,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'uvicorn[standard]>=0.13.3,<0.14.0']

setup_kwargs = {
    'name': 'coresvc',
    'version': '0.1.7',
    'description': 'core services in microservice architecture',
    'long_description': None,
    'author': 'MartinJ-Dev',
    'author_email': '29751536+MartinJ-Dev@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
